__author__ = 'Kemele M. Endris and Philipp D. Rohde'

import urllib.parse
import urllib.request
from multiprocessing import Process, Queue

import DeTrusty.Decomposer.utils as utils
from DeTrusty.Decomposer.Tree import Leaf, Node
from DeTrusty.Operators.AnapsidOperators.Xbind import Xbind
from DeTrusty.Operators.AnapsidOperators.Xdistinct import Xdistinct
from DeTrusty.Operators.AnapsidOperators.Xfilter import Xfilter
from DeTrusty.Operators.AnapsidOperators.Xgjoin import Xgjoin
from DeTrusty.Operators.AnapsidOperators.Xgoptional import Xgoptional
from DeTrusty.Operators.AnapsidOperators.Xgroupby import Xgroupby
from DeTrusty.Operators.AnapsidOperators.Xhaving import Xhaving
from DeTrusty.Operators.AnapsidOperators.Xlimit import Xlimit
from DeTrusty.Operators.AnapsidOperators.Xoffset import Xoffset
from DeTrusty.Operators.AnapsidOperators.Xorderby import Xorderby
from DeTrusty.Operators.AnapsidOperators.Xproject import Xproject
from DeTrusty.Operators.AnapsidOperators.Xunion import Xunion
from DeTrusty.Operators.BlockingOperators.Union import Union
from DeTrusty.Operators.NonBlockingOperators.NestedHashJoinFilter import NestedHashJoinFilter as NestedHashJoin
from DeTrusty.Operators.NonBlockingOperators.NestedHashOptionalFilter import  NestedHashOptionalFilter as NestedHashOptional
from DeTrusty.Sparql.Parser.services import Bind, Filter, Service, Optional, UnionBlock, JoinBlock


class Planner(object):

    def __init__(self, query,  wc, contact, endpointType, config):
        self.query = query
        self.withcounts = wc
        self.contact = contact
        self.endpType = endpointType
        self.config = config
        self.buffersize = 16384
        self.adaptive = True

    def createPlan(self):
        if self.query is None:
            return None

        query = self.query
        operatorTree = self.includePhysicalOperatorsQuery()
        aggs = list()

        over_all_triples = utils.collectVars(query.args, query.group_by, query.having, aggs)  # TODO: test & fix
        # print(query.args, query.group_by, query.having, aggs, over_all_triples)

        # Adds the group by operator to the plan.
        if (len(query.group_by) > 0):
            operatorTree = TreePlan(Xgroupby(query.group_by), operatorTree.vars, operatorTree)

        # Adds the group by operator to the plan.
        # if aggs:
            # operatorTree = TreePlan(Xaggregate(aggs, over_all_triples, query.group_by, query.prefs), operatorTree.vars, operatorTree)

        # Adds the having operator to the plan.
        if query.having is not None:
            operatorTree = TreePlan(Xhaving(query.having), operatorTree.vars, operatorTree)

        # Adds the order by operator to the plan.
        if (len(query.order_by) > 0):
            operatorTree = TreePlan(Xorderby(query.order_by), operatorTree.vars, operatorTree)

        # Adds the project operator to the plan.
        operatorTree = TreePlan(Xproject(query.args), operatorTree.vars, operatorTree)

        # Adds the distinct operator to the plan.
        if (query.distinct):
            operatorTree = TreePlan(Xdistinct(None), operatorTree.vars, operatorTree)

        # Adds the offset operator to the plan.
        if (query.offset != -1):
            operatorTree = TreePlan(Xoffset(None, query.offset), operatorTree.vars, operatorTree)

        # Adds the limit operator to the plan.
        if (query.limit != -1):
            # print "query.limit", query.limit
            operatorTree = TreePlan(Xlimit(None, query.limit), operatorTree.vars, operatorTree)

        return operatorTree

    def includePhysicalOperatorsQuery(self):
        return self.includePhysicalOperatorsUnionBlock(self.query.body)

    def includePhysicalOperatorsUnionBlock(self, ub):
        r = []
        for jb in ub.triples:
            pp = self.includePhysicalOperatorsJoinBlock(jb)
            r.append(pp)

        while len(r) > 1:
            left = r.pop(0)
            right = r.pop(0)
            all_variables = left.vars | right.vars
            if self.adaptive:
                n = TreePlan(Xunion(left.vars, right.vars), all_variables, left, right)
            else:
                n = TreePlan(Union(left.vars, right.vars, self.query.distinct), all_variables, left, right)
            r.append(n)

        if len(r) == 1:
            n = r[0]
            for f in ub.filters:
                n = TreePlan(Xfilter(f), n.vars, n)
            return n
        else:
            return None

    def includePhysicalOperatorsJoinBlock(self, jb):
        tl = []
        ol = []
        if isinstance(jb.triples, list):
            for bgp in jb.triples:
                if isinstance(bgp, Node) or isinstance(bgp, Leaf):
                    tl.append(self.includePhysicalOperators(bgp))
                elif isinstance(bgp, Optional):
                    ol.append(self.includePhysicalOperatorsUnionBlock(bgp.bgg))
                elif isinstance(bgp, UnionBlock):
                    tl.append(self.includePhysicalOperatorsUnionBlock(bgp))
        elif isinstance(jb.triples, Node) or isinstance(jb.triples, Leaf):
            tl = [self.includePhysicalOperators(jb.triples)]
        else:  # this should never be the case..
            pass

        while len(tl) > 1:
            l = tl.pop(0)
            r = tl.pop(0)
            n = self.includePhysicalOperatorJoin(l, r)
            tl.append(n)

        if len(tl) == 1:
            nf = self.includePhysicalOperatorsOptional(tl[0], ol)
            if isinstance(tl[0], TreePlan) and (isinstance(tl[0].operator, Xfilter) or isinstance(tl[0].operator, Xbind)):
                return nf
            else:
                if len(jb.filters) > 0 and isinstance(tl[0].operator, Xfilter):
                    for f in jb.filters:
                        nf = TreePlan(Xfilter(f), nf.vars, nf)
                if len(jb.filters) > 0 and isinstance(tl[0].operator, Xbind):
                    for f in jb.filters:
                        nf = TreePlan(Xbind(f), nf.vars, nf)
                    return nf
                else:
                    return nf
        else:
            return None

    def includePhysicalOperators(self, tree):
        if isinstance(tree, Leaf):
            if isinstance(tree.service, Service):
                if tree.filters == []:
                    return IndependentOperator(self.query, tree, self.contact, self.config)
                else:
                    n = IndependentOperator(self.query, tree, self.contact, self.config)
                    for f in tree.filters:
                        vars_f = f.getVarsName()
                        if set(n.vars) & set(vars_f) == set(vars_f):
                            n = TreePlan(Xfilter(f), n.vars, n)
                    return n
            elif isinstance(tree.service, UnionBlock):
                return self.includePhysicalOperatorsUnionBlock(tree.service)
            elif isinstance(tree.service, JoinBlock):
                if tree.filters == []:
                    return self.includePhysicalOperatorsJoinBlock(tree.service)
                else:
                    n = self.includePhysicalOperatorsJoinBlock(tree.service)
                    for f in tree.filters:
                        vars_f = f.getVarsName()
                        if set(n.vars) & set(vars_f) == set(vars_f):
                            n = TreePlan(Xfilter(f), n.vars, n)
                    return n
            else:
                print("tree.service" + str(type(tree.service)) + str(tree.service))
                print("Error Type not considered")

        elif isinstance(tree, Node):
            left_subtree = self.includePhysicalOperators(tree.left)
            right_subtree = self.includePhysicalOperators(tree.right)
            if tree.filters == []:
                return self.includePhysicalOperatorJoin(left_subtree, right_subtree)
            else:
                n = self.includePhysicalOperatorJoin(left_subtree, right_subtree)
                for f in tree.filters:
                    vars_f = f.getVarsName()
                    if set(n.vars) & set(vars_f) == set(vars_f):
                        if isinstance(f, Filter):
                            n = TreePlan(Xfilter(f), n.vars, n)
                    if isinstance(f, Bind):
                        n.vars = set(n.vars) | set(vars_f)
                        if n.vars & set(vars_f) == set(vars_f):
                            n = TreePlan(Xbind(f), n.vars, n)
            return n

    def includePhysicalOperatorJoinX(self, l, r):
        join_variables = l.vars & r.vars
        all_variables = l.vars | r.vars
        n, dependent_join = self.joinIndependentAnapsid(l, r)
        if n and isinstance(n.left, IndependentOperator) and isinstance(n.left.tree, Leaf):
            if (n.left.constantPercentage() <= 0.5) and not (n.left.tree.service.allTriplesGeneral()):
                n.left.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
        # elif not decided:
            # n = TreePlan(Xgjoin(join_variables), all_variables, l, r)
        # print "n: ", n
        if isinstance(n.right, IndependentOperator) and isinstance(n.right.tree, Leaf):
            if not dependent_join:
                if (n.right.constantPercentage() <= 0.5) and not (n.right.tree.service.allTriplesGeneral()):
                    n.right.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
                    # print "modifying limit right ..."
            else:
                new_constants = 0
                for v in join_variables:
                    new_constants = new_constants + n.right.query.show().count(v)
                if ((n.right.constantNumber() + new_constants) / n.right.places() <= 0.5) and not (n.right.tree.service.allTriplesGeneral()):
                    n.right.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
        return n

    def includePhysicalOperatorJoin(self, l, r):
        join_variables = l.vars & r.vars
        all_variables = l.vars | r.vars
        # noInstantiatedLeftStar = False
        # noInstantiatedRightStar = False
        lowSelectivityLeft = l.allTriplesLowSelectivity()
        lowSelectivityRight = r.allTriplesLowSelectivity()
        n = None
        dependent_join = False

        if isinstance(l, IndependentOperator) and isinstance(l.tree.service.triples, list) and l.tree.service.triples[0].subject.constant:
            if len(join_variables) > 0:
                n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                dependent_join = True
        elif isinstance(r, IndependentOperator) and isinstance(r.tree.service.triples, list) and r.tree.service.triples[0].subject.constant:
            if len(join_variables) > 0:
                n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                dependent_join = True
        elif isinstance(r, TreePlan) and r.operator.__class__.__name__ == "Xunion" and \
                isinstance(l, TreePlan) and l.operator.__class__.__name__ == "Xunion":
            # both are Union operators
            n = TreePlan(Xgjoin(join_variables), all_variables, l, r)

        elif not lowSelectivityLeft and not lowSelectivityRight and (
                not isinstance(l, TreePlan) or not isinstance(r, TreePlan)):
            # if both are selective and one of them (or both) are Independent Operator
            if len(join_variables) > 0:
                if l.constantPercentage() > r.constantPercentage():
                    if not isinstance(r, TreePlan):
                        n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                        dependent_join = True
                    else:
                        n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                else:
                    if not isinstance(l, TreePlan):
                        n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                        dependent_join = True
                    else:
                        n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)

        elif not lowSelectivityLeft and lowSelectivityRight and not isinstance(r, TreePlan):

            # If left is selective, if left != NHJ and right != NHJ -> NHJ (l,r)
            if len(join_variables) > 0:
                n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                dependent_join = True

                # if not isinstance(l, TreePlan):
                #     if isinstance(r, TreePlan):
                #         if not isinstance(r.operator, NestedHashJoin) and not isinstance(r.operator, Xgjoin):
                #             n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                #             dependent_join = True
                #     else:
                #         # IF both are Independent Operators
                #         n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                #         dependent_join = True
                # elif isinstance(l, TreePlan) and not isinstance(l.operator, NestedHashJoin):
                #     if not isinstance(r, TreePlan):
                #         n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                #         dependent_join = True
                #     elif isinstance(r, TreePlan) and not isinstance(r.operator, NestedHashJoin) and not isinstance(r.operator, Xgjoin):
                #         n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                #         dependent_join = True

        elif lowSelectivityLeft and not lowSelectivityRight and not isinstance(l, TreePlan):
            # if right is selective if left != NHJ and right != NHJ -> NHJ (r,l)
            if len(join_variables) > 0:
                n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                dependent_join = True

                # if not isinstance(r, TreePlan):
                #     if isinstance(l, TreePlan):
                #         if not isinstance(l.operator, NestedHashJoin) and not isinstance(l.operator, Xgjoin):
                #             n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                #             dependent_join = True
                #     else:
                #         # IF both are Independent Operators
                #         n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                #         dependent_join = True
                # elif isinstance(r, TreePlan) and not isinstance(r.operator, NestedHashJoin):
                #     if not isinstance(l, TreePlan):
                #         n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                #         dependent_join = True
                #     elif isinstance(l, TreePlan) and not isinstance(l.operator, NestedHashJoin) and not isinstance(l.operator, Xgjoin):
                #         n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                #         dependent_join = True
        elif not lowSelectivityLeft and lowSelectivityRight \
                and (isinstance(l, TreePlan) and not l.operator.__class__.__name__ == "NestedHashJoinFilter") \
                and (isinstance(r, TreePlan) and not (r.operator.__class__.__name__ == "NestedHashJoinFilter"
                                                      or r.operator.__class__.__name__ == "Xgjoin")):
            if len(join_variables) > 0:
                n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                dependent_join = True
        elif lowSelectivityLeft and not lowSelectivityRight and (isinstance(r, TreePlan) and not r.operator.__class__.__name__ == "NestedHashJoinFilter") \
                and (isinstance(l, TreePlan) and not (l.operator.__class__.__name__ == "NestedHashJoinFilter" or l.operator.__class__.__name__ == "Xgjoin")):
            if len(join_variables) > 0:
                n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                dependent_join = True
        elif lowSelectivityLeft and lowSelectivityRight and isinstance(l, IndependentOperator) and isinstance(r, IndependentOperator):
            # both are non-selective and both are Independent Operators
            n = TreePlan(Xgjoin(join_variables), all_variables, r, l)

        if n is None:
            n = TreePlan(Xgjoin(join_variables), all_variables, l, r)

        if n and isinstance(n.left, IndependentOperator) and isinstance(n.left.tree, Leaf):
            if (n.left.constantPercentage() <= 0.5) and not (n.left.tree.service.allTriplesGeneral()):
                n.left.tree.service.limit = 10000  # Fixed value, this can be learnt in the future

        if isinstance(n.right, IndependentOperator) and isinstance(n.right.tree, Leaf):
            if not dependent_join:
                if (n.right.constantPercentage() <= 0.5) and not (n.right.tree.service.allTriplesGeneral()):
                    n.right.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
                    # print "modifying limit right ..."
            else:
                new_constants = 0
                for v in join_variables:
                    new_constants = new_constants + n.right.query.show().count(v)
                if ((n.right.constantNumber() + new_constants) / n.right.places() <= 0.5) and not (n.right.tree.service.allTriplesGeneral()):
                    n.right.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
        return n

    def includePhysicalOperatorsOptional(self, left, rightList):
        l = left
        for right in rightList:
            all_variables = l.vars | right.vars
            lowSelectivityLeft = l.allTriplesLowSelectivity()
            lowSelectivityRight = right.allTriplesLowSelectivity()
            join_variables = l.vars & right.vars
            dependent_op = False
            # if left.operator.__class__.__name__ == "NestedHashJoinFilter":
            #     l = TreePlan(Xgoptional(left.vars, right.vars), all_variables, l, right)
            # Case 1: left operator is highly selective and right operator is low selective
            if not (lowSelectivityLeft) and lowSelectivityRight and not (isinstance(right, TreePlan)):
                l = TreePlan(NestedHashOptional(l.vars, right.vars), all_variables, l, right)
                dependent_op = True

            # Case 2: left operator is low selective and right operator is highly selective
            elif lowSelectivityLeft and not (lowSelectivityRight) and not (isinstance(right, TreePlan)):
                l = TreePlan(NestedHashOptional(l.vars, right.vars), all_variables, right, l)
                dependent_op = True

            elif not lowSelectivityLeft and lowSelectivityRight and not (isinstance(left, TreePlan) and (left.operator.__class__.__name__ == "NestedHashJoinFilter" or left.operator.__class__.__name__ == "Xgjoin")) \
                    and not (isinstance(right, IndependentOperator)) \
                    and not (right.operator.__class__.__name__ == "NestedHashJoinFilter" or right.operator.__class__.__name__ == "Xgjoin") \
                    and (right.operator.__class__.__name__ == "Xunion"):
                l = TreePlan(NestedHashOptional(l.vars, right.vars), all_variables, l, right)
                dependent_op = True
            # Case 3: both operators are low selective
            else:
                l = TreePlan(Xgoptional(l.vars, right.vars), all_variables, l, right)
                # print "Planner CASE 3: xgoptional"

            if isinstance(l.left, IndependentOperator) and isinstance(l.left.tree, Leaf) and not l.left.tree.service.allTriplesGeneral():
                if l.left.constantPercentage() <= 0.5:
                    l.left.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
                    # print "modifying limit optional left ..."

            if isinstance(l.right, IndependentOperator) and isinstance(l.right.tree, Leaf):
                if not dependent_op:
                    if (l.right.constantPercentage() <= 0.5) and not (l.right.tree.service.allTriplesGeneral()):
                        l.right.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
                        # print "modifying limit optional right ..."
                else:
                    new_constants = 0
                    for v in join_variables:
                        new_constants = new_constants + l.right.query.show().count(v)
                    if ((l.right.constantNumber() + new_constants) / l.right.places() <= 0.5) and not l.right.tree.service.allTriplesGeneral():
                        l.right.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
                        # print "modifying limit optional right ..."

        return l

    def joinIndependentAnapsid(self, l, r):
        join_variables = l.vars & r.vars
        all_variables = l.vars | r.vars
        noInstantiatedLeftStar = False
        noInstantiatedRightStar = False
        lowSelectivityLeft = l.allTriplesLowSelectivity()
        lowSelectivityRight = r.allTriplesLowSelectivity()

        dependent_join = False
        # if (noInstantiatedRightStar) or ((not wc) and (l.constantPercentage() >= 0.5) and (len(join_variables) > 0) and c):
        # Case 1: left operator is highly selective and right operator is low selective
        if not (lowSelectivityLeft) and lowSelectivityRight and not (isinstance(r, TreePlan)):
            n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
            dependent_join = True
            # print "Planner CASE 1: nested loop", type(r)
        # Case 2: left operator is low selective and right operator is highly selective
        elif lowSelectivityLeft and not (lowSelectivityRight) and not (isinstance(l, TreePlan)):
            n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
            dependent_join = True
            # print "Planner CASE 2: nested loop swapping plan", type(r)
        elif not (lowSelectivityLeft) and lowSelectivityRight and (
                not (isinstance(l, TreePlan)) or not (l.operator.__class__.__name__ == "NestedHashJoinFilter")) and (
                not (isinstance(r, TreePlan)) or not (
                r.operator.__class__.__name__ == "Xgjoin" or r.operator.__class__.__name__ == "NestedHashJoinFilter")):
            if (isinstance(r, TreePlan) and (set(l.vars) & set(r.operator.vars_left) != set([])) and (
                    set(l.vars) & set(r.operator.vars_right) != set([]))):
                n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                dependent_join = True
            elif (isinstance(l, TreePlan) and (set(r.vars) & set(l.operator.vars_left) != set([])) and (
                    set(r.vars) & set(l.operator.vars_right) != set([]))):
                n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                dependent_join = True
            else:
                n = TreePlan(Xgjoin(join_variables), all_variables, l, r)
            # print "Planner case 2.5", type(r)
        # Case 3: both operators are low selective

        else:
            n = TreePlan(Xgjoin(join_variables), all_variables, l, r)
            # print "Planner CASE 3: xgjoin"

        if isinstance(n.left, IndependentOperator) and isinstance(n.left.tree, Leaf):
            if (n.left.constantPercentage() <= 0.5) and not (n.left.tree.service.allTriplesGeneral()):
                n.left.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
                # print "modifying limit left ..."

        if isinstance(n.right, IndependentOperator) and isinstance(n.right.tree, Leaf):
            if not (dependent_join):
                if (n.right.constantPercentage() <= 0.5) and not (n.right.tree.service.allTriplesGeneral()):
                    n.right.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
                    # print "modifying limit right ..."
            else:
                new_constants = 0
                for v in join_variables:
                    new_constants = new_constants + n.right.query.show().count(v)
                if ((n.right.constantNumber() + new_constants) / n.right.places() <= 0.5) and not (
                n.right.tree.service.allTriplesGeneral()):
                    n.right.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
                    # print "modifying limit right ..."

        return n, dependent_join

    def joinIndependentMULDER(self, l, r):
        join_variables = l.vars & r.vars
        all_variables = l.vars | r.vars
        # noInstantiatedLeftStar = False
        # noInstantiatedRightStar = False
        lowSelectivityLeft = l.allTriplesLowSelectivity()
        lowSelectivityRight = r.allTriplesLowSelectivity()
        n = None
        dependent_join = False

        if isinstance(l, IndependentOperator) and l.tree.service.triples[0].subject.constant:
            if len(join_variables) > 0:
                n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                dependent_join = True
        elif isinstance(r, IndependentOperator) and r.tree.service.triples[0].subject.constant:
            if len(join_variables) > 0:
                n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                dependent_join = True
        elif isinstance(r, TreePlan) and r.operator.__class__.__name__ == "Xunion" and \
                isinstance(l, TreePlan) and l.operator.__class__.__name__ == "Xunion":
            # both are Union operators
            n = TreePlan(Xgjoin(join_variables), all_variables, l, r)
        elif not lowSelectivityLeft and not lowSelectivityRight and (
                not isinstance(l, TreePlan) or not isinstance(r, TreePlan)):
            # if both are selective and one of them (or both) are Independent Operator
            if len(join_variables) > 0:
                if l.constantPercentage() > r.constantPercentage():
                    if not isinstance(r, TreePlan):
                        n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                        dependent_join = True
                    else:
                        n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                else:
                    if not isinstance(l, TreePlan):
                        n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                        dependent_join = True
                    else:
                        n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)

        elif not lowSelectivityLeft and lowSelectivityRight and not isinstance(r, TreePlan):

            # If left is selective, if left != NHJ and right != NHJ -> NHJ (l,r)
            if len(join_variables) > 0:
                n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                dependent_join = True
        elif lowSelectivityLeft and not lowSelectivityRight and not isinstance(l, TreePlan):
            # if right is selective if left != NHJ and right != NHJ -> NHJ (r,l)
            if len(join_variables) > 0:
                n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                dependent_join = True
        elif not lowSelectivityLeft and lowSelectivityRight and not l.__class__.__name__ == "NestedHashJoinFilter" and not (
                r.__class__.__name__ == "NestedHashJoinFilter" or r.__class__.__name__ == "Xgjoin"):
            if len(join_variables) > 0:
                n = TreePlan(NestedHashJoin(join_variables), all_variables, l, r)
                dependent_join = True
        elif lowSelectivityLeft and not lowSelectivityRight and not r.__class__.__name__ == "NestedHashJoinFilter" and not (
                l.__class__.__name__ == "NestedHashJoinFilter" or l.__class__.__name__ == "Xgjoin"):
            if len(join_variables) > 0:
                n = TreePlan(NestedHashJoin(join_variables), all_variables, r, l)
                dependent_join = True
        elif lowSelectivityLeft and lowSelectivityRight and isinstance(l, IndependentOperator) and isinstance(r, IndependentOperator):
            # both are non-selective and both are Independent Operators
            n = TreePlan(Xgjoin(join_variables), all_variables, r, l)

        if n is None:
            n = TreePlan(Xgjoin(join_variables), all_variables, l, r)

        if isinstance(n.left, IndependentOperator) and isinstance(n.left.tree, Leaf):
            if (n.left.constantPercentage() <= 0.5) and not (n.left.tree.service.allTriplesGeneral()):
                n.left.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
                # print "modifying limit left ..."

        if isinstance(n.right, IndependentOperator) and isinstance(n.right.tree, Leaf):
            if not (dependent_join):
                if (n.right.constantPercentage() <= 0.5) and not (n.right.tree.service.allTriplesGeneral()):
                    n.right.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
                    # print "modifying limit right ..."
            else:
                new_constants = 0
                for v in join_variables:
                    new_constants = new_constants + n.right.query.show().count(v)
                if ((n.right.constantNumber() + new_constants) / n.right.places() <= 0.5) and not (
                n.right.tree.service.allTriplesGeneral()):
                    n.right.tree.service.limit = 10000  # Fixed value, this can be learnt in the future
                    # print "modifying limit right ..."

        return n, dependent_join


class TreePlan(object):
    """
    Represents a plan to be executed by the engine.

    It is composed by a left node, a right node, and an operator node.
    The left and right nodes can be leaves to contact sources, or subtrees.
    The operator node is a physical operator, provided by the engine.

    The execute() method evaluates the plan.
    It creates a process for every node of the plan.
    The left node is always evaluated.
    If the right node is an independent operator or a subtree, it is evaluated.
    """
    def __init__(self, operator, vars, left=None, right=None):
        self.operator = operator
        self.vars = vars
        self.left = left
        self.right = right
        self.cardinality = None
        self.joinCardinality = []

    def __repr__(self):
        return self.aux(" ")

    def instantiate(self, d):
        l = None
        r = None
        if self.left:
            l = self.left.instantiate(d)
        if self.right:
            r = self.right.instantiate(d)
        newvars = self.vars - set(d.keys())

        return TreePlan(self.operator.instantiate(d), newvars, l, r)

    def instantiateFilter(self, d, filter_str):
        l = None
        r = None
        if self.left:
            l = self.left.instantiateFilter(d, filter_str)
        if self.right:
            r = self.right.instantiateFilter(d, filter_str)
        newvars = self.vars - set(d)
        return TreePlan(self.operator.instantiateFilter(d, filter_str), newvars, l, r)

    def allTriplesLowSelectivity(self):
        a = True
        if self.left:
            a = self.left.allTriplesLowSelectivity()
        if self.right:
            a = a and self.right.allTriplesLowSelectivity()
        return a

    def places(self):
        p = 0
        if self.left:
            p = self.left.places()
        if self.right:
            p = p + self.right.places()
        return p

    def constantNumber(self):
        c = 0
        if self.left:
            c = self.left.constantNumber()
        if self.right:
            c = c + self.right.constantNumber()
        return c

    def constantPercentage(self):
        return self.constantNumber()/self.places()

    def getCardinality(self):
        if self.cardinality is None:
            self.cardinality = self.operator.getCardinality(self.left, self.right)
        return self.cardinality

    def getJoinCardinality(self, vars):
        c = None
        for (v, c2) in self.joinCardinality:
            if v == vars:
                c = c2
                break
        if c is None:
            c = self.operator.getJoinCardinality(self.left, self.right, vars)
            self.joinCardinality.append((vars, c))
        return c

    def aux(self, n):
        s = n + str(self.operator) + "\n" + n + str(self.vars) + "\n"
        if self.left:
            s = s + self.left.aux(n+"  ")

        if self.right:
            s = s + self.right.aux(n + "  ")
        return s

    def json(self, triples=None, sub_queries=None):
        if triples is None:
            triples = {}
        if sub_queries is None:
            sub_queries = [0]
        tree = {'name': self.operator.name}
        children = []
        if self.left:
            child, _ = self.left.json(triples, sub_queries)
            children.append(child)
        if self.right:
            child, _ = self.right.json(triples, sub_queries)
            children.append(child)
        if children:
            tree['children'] = children
        return tree, triples

    def execute(self, outputqueue, processqueue=Queue()):
        # Evaluates the execution plan.
        if self.left:  # and this.right: # This line was modified by mac in order to evaluate unary operators
            qleft  = Queue()
            qright = Queue()

            # The left node is always evaluated.
            # Create process for left node
            # print("self.right:", self.right)
            # print("self.left:", self.left)

            p1 = Process(target=self.left.execute, args=(qleft, processqueue, ))
            p1.start()
            # processqueue.put(p1.pid)
            if "Nested" in self.operator.__class__.__name__:
                p3 = Process(target=self.operator.execute, args=(qleft, self.right, outputqueue, processqueue, ))
                p3.start()
                # processqueue.put(p3.pid)
                return

            # Check the right node to determine if evaluate it or not.
            if self.right and ((self.right.__class__.__name__ == "IndependentOperator") or (self.right.__class__.__name__ == "TreePlan")):
                p2 = Process(target=self.right.execute, args=(qright, processqueue, ))
                p2.start()
                # processqueue.put(p2.pid)
            else:
                qright = self.right  # qright.put("EOF")

            # Create a process for the operator node.
            p = Process(target=self.operator.execute, args=(qleft, qright, outputqueue, processqueue, ))
            # print("left and right")
            # Execute the plan
            p.start()
            # processqueue.put(p.pid)


class IndependentOperator(object):
    """
    Implements an operator that can be resolved independently.

    It receives as input the url of the server to be contacted, the
    filename that contains the query, the header size of the messages.

    The execute() method reads tuples from the input queue and
    response message and the buffer size (length of the string)
    place them in the output queue.
    """
    def __init__(self, query, tree, c, config):
        (e, sq, vs) = tree.getInfoIO(query)
        self.contact = c
        self.server = e
        self.query = query
        self.tree = tree
        self.query_str = sq
        self.vars = vs
        self.buffersize = 163840
        self.config = config
        self.cardinality = None
        self.joinCardinality = []
        # self.limit = limit

    def __repr__(self):
        return str(self.tree)

    def json(self, triples, sub_queries):
        sub_queries[0] += 1
        name = 'SSQ' + str(sub_queries[0])

        if isinstance(self.tree.service.triples, list):
            triples_str = '\n'.join([str(x).strip() for x in self.tree.service.triples])
        else:
            triples_str = str(self.tree.service.triples)
        triples_str = triples_str.replace('<', '&lt;').replace('>', '&gt;')

        triples[name] = {
            'endpoint': self.server,
            'triples': triples_str
        }
        return {'name': name, 'endpoint': self.server}, None

    def instantiate(self, d):
        new_tree = self.tree.instantiate(d)
        return IndependentOperator(self.query, new_tree, self.contact, self.config)

    def instantiateFilter(self, vars_instantiated, filter_str):
        new_tree = self.tree.instantiateFilter(vars_instantiated, filter_str)
        return IndependentOperator(self.query, new_tree, self.contact, self.config)

    def getCardinality(self):
        if self.cardinality is None:
            self.cardinality = self.askCount(self.query, self.tree, set(), self.contact)
        return self.cardinality

    def askCount(self, query, tree, vars, contact):
        (server, query) = tree.getCount(query, vars, None)
        q = Queue()
        contact(server, query, q)

        res = q.get()
        # print res
        v = -1
        if res == "EOF":
            return 20000
        for k in res:
            v = res[k]
        q.get()
        return int(v)

    def getJoinCardinality(self, vars):
        c = None
        for (v, c2) in self.joinCardinality:
            if v == vars:
                c = c2
                break
        if c is None:
            if len(vars) == 0:
                c = self.getCardinality()
            else:
                c = self.askCount(self.query, self.tree, vars, self.contact)
            self.joinCardinality.append((vars, c))
        return c

    def allTriplesLowSelectivity(self):
        return self.tree.service.allTriplesLowSelectivity()

    def places(self):
        return self.tree.places()

    def constantNumber(self):

        return self.tree.constantNumber()

    def constantPercentage(self):
        return self.constantNumber()/self.places()

    def aux(self, n):
        return self.tree.aux(n)

    def execute(self, outputqueue, processqueue=Queue()):

        if self.tree.service.limit == -1:
            self.tree.service.limit = 10000  # TODO: Fixed value, this can be learnt in the future

        # Evaluate the independent operator.

        # self.q = Queue()

        p = Process(target=self.contact, args=(self.server, self.query_str, outputqueue, self.config, self.tree.service.limit))
        p.start()
        # processqueue.put(p.pid)

        # i = 0
        # while True:
        #     # Get the next item in queue.
        #     res = self.q.get(True)
        #     # Put the result into the output queue.
        #     #print res
        #     i += 1
        #     outputqueue.put(res)
        #     # Check if there's no more data.
        #     if res == "EOF":
        #         break

        # p.terminate()


def contactSource(molecule, query, queue, config, limit=-1):
    # Contacts the datasource (i.e. real endpoint).
    # Every tuple in the answer is represented as Python dictionaries
    # and is stored in a queue.
    # print "in *NEW* contactSource"
    b = None
    cardinality = 0
    # moleculeMetadata = config.findMolecule(molecule)
    # wrappers = [w for w in moleculeMetadata['wrappers']]
    # if len(wrappers) == 0:
    #     queue.put("EOF")
    # server = wrappers[0]['url']
    server = molecule

    referer = server
    server = server.split("http://")[1]
    if '/' in server:
        (server, path) = server.split("/", 1)
    else:
        path = ""
    host_port = server.split(":")
    port = 80 if len(host_port) == 1 else host_port[1]
    card = 0
    if limit == -1:
        b, card = contactSourceAux(referer, server, path, port, query, queue)
    else:
        # Contacts the datasource (i.e. real endpoint) incrementally,
        # retreiving partial result sets combining the SPARQL sequence
        # modifiers LIMIT and OFFSET.

        # Set up the offset.
        offset = 0

        while True:
            query_copy = query + " LIMIT " + str(limit) + " OFFSET " + str(offset)
            b, cardinality = contactSourceAux(referer, server, path, port, query_copy, queue)
            card += cardinality
            if cardinality < limit:
                break

            offset = offset + limit

    # Close the queue
    queue.put("EOF")
    return b


def contactSourceAux(referer, server, path, port, query, queue):
    # Setting variables to return.
    b = None
    reslist = 0

    if '0.0.0.0' in server:
        server = server.replace('0.0.0.0', 'localhost')

    js = "application/sparql-results+json"
    params = {'query': query, 'format': js}
    headers = {"User-Agent":
                   "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.77 Safari/537.36",
               "Accept": js}
    try:
        data = urllib.parse.urlencode(params)
        data = data.encode('utf-8')
        req = urllib.request.Request(referer, data, headers)
        with urllib.request.urlopen(req) as response:
            resp = response.read()
            resp = resp.decode()
            res = resp.replace("false", "False")
            res = res.replace("true", "True")
            res = eval(res)
            reslist = 0
            if type(res) == dict:
                b = res.get('boolean', None)

                if 'results' in res:
                    # print "raw results from endpoint", res
                    for x in res['results']['bindings']:
                        for key, props in x.items():
                            # Handle typed-literals and language tags
                            suffix = ''
                            if props['type'] == 'typed-literal':
                                if isinstance(props['datatype'], bytes):
                                    suffix = "^^<" + props['datatype'].decode('utf-8') + ">"
                                else:
                                    suffix = "^^<" + props['datatype'] + ">"
                            elif "xml:lang" in props:
                                suffix = '@' + props['xml:lang']
                            try:
                                if isinstance(props['value'], bytes):
                                    x[key] = props['value'].decode('utf-8') + suffix
                                else:
                                    x[key] = props['value'] + suffix
                            except:
                                x[key] = props['value'] + suffix

                        queue.put(x)
                        reslist += 1
                    # Every tuple is added to the queue.
                    # for elem in reslist:
                        # print elem
                        # queue.put(elem)

            else:
                print("the source " + str(server) + " answered in " + res.getheader("content-type") + " format, instead of"
                       + " the JSON format required, then that answer will be ignored")
    except Exception as e:
        print("Exception while sending request to ", referer, "msg:", e)

    # print "b - ", b
    # print server, query, len(reslist)

    # print "Contact Source returned: ", len(reslist), ' results'
    return b, reslist



from multiprocessing import Queue
import datetime

data_types = {
    'integer': (int, 'numerical'),
    'decimal': (float, 'numerical'),
    'float': (float, 'numerical'),
    'double': (float, 'numerical'),
    'string': (str, str),
    'boolean': (bool, bool),
    'dateTime': (datetime, datetime),
    'nonPositiveInteger': (int, 'numerical'),
    'negativeInteger': (int, 'numerical'),
    'long': (int, 'numerical'),
    'int': (int, 'numerical'),
    'short': (int, 'numerical'),
    'byte': (bytes, bytes),
    'nonNegativeInteger': (int, 'numerical'),
    'unsignedLong': (int, 'numerical'),
    'unsignedInt': (int, 'numerical'),
    'unsignedShort': (int, 'numerical'),
    'unsignedByte': (bytes, bytes),  # TODO: this is not correct
    'positiveInteger': (int, 'numerical')
}


class Xvalues(object):

    name = "VALUES"

    def __init__(self, values): 
        self.input = Queue()
        self.qresults = Queue()
        self.values = values  

    def execute(self, left, dummy, out, processqueue=Queue()):
        self.left = left
        self.qresults = out
        tuple = self.left.get(True)

        while (tuple != "EOF"):
            shouldInclude = self.filterByValues(tuple)
            if shouldInclude:
                self.qresults.put(tuple)

            tuple = self.left.get(True)           

        # Put EOF in queue and exit.
        self.qresults.put("EOF")

    def filterByValues(self, tuple):        
        for row in self.values.data_block_val:
            isValid = True
            for idx, variable in enumerate(self.values.var): 
                value = tuple[variable.name[1:]]
                rowArg = row[idx]
                if rowArg is None:
                    continue 
                extractedRowValue = self.extractValue(rowArg.name)
                rowValue = extractedRowValue[1:-1]
                
                if value != rowValue:
                    isValid = False
                    break
            if isValid:
                return True
        return False 

    def extractValue(self, val):
        pos = val.find("^^")
        # Handles when the literal is typed.
        if (pos > -1):
            for t in data_types.keys():
                if (t in val[pos:]):
                    (python_type, general_type) = data_types[t]
                    if (general_type == bool):
                        return val[:pos]
                    else:
                        return python_type(val[:pos])
        else:
            return str(val)



#!/usr/bin/env python3

import getopt
import json
import pathlib
import sys

PACKAGE_DETRUSTY = str(pathlib.Path(__file__).parent.parent.resolve())
sys.path.append(PACKAGE_DETRUSTY)
from DeTrusty.Molecule.MTCreation import DEFAULT_OUTPUT_PATH, create_rdfmts, logger
sys.path.remove(PACKAGE_DETRUSTY)


def get_options(argv):
    try:
        opts, args = getopt.getopt(argv, 'h:s:o:j:i')
    except getopt.GetoptError:
        usage()
        sys.exit(1)

    endpoints_file = None
    output = DEFAULT_OUTPUT_PATH
    is_json = False
    interlinking = False
    for opt, arg in opts:
        if opt == '-h':
            usage()
            sys.exit()
        elif opt == '-s':
            endpoints_file = arg
        elif opt == '-o':
            output = arg
        elif opt == '-j':
            is_json = True
        elif opt == '-i':
            interlinking = True

    if not endpoints_file:
        usage()
        sys.exit(1)

    if not is_json:
        with open(endpoints_file, 'r') as f:
            endpoints = f.readlines()
            if len(endpoints) == 0:
                logger.critical("The endpoints file should contain at least one URL")
                sys.exit(1)
            endpoints = [e.strip('\n') for e in endpoints]
    else:
        with open(endpoints_file, 'r') as f:
            endpoints = json.load(f)

    if '.json' not in output:
        output += '.json'
    return endpoints, output, interlinking


def usage():
    usage_str = (
        'Usage: {program} -s <path/to/endpoints.txt> [-o <path/to/output.json>] [-j] [-i]\n'
        'where\n'
        '    <path/to/endpoints.txt> - path to a text file containing a list of SPARQL endpoint URLs\n'
        '    <path/to/output.json> - name of output file\n'
        'parameters\n'
        '    -j\tif set, the endpoints file will be handled as JSON instead of plain text\n'
        '    -i\tif set, interlinks between the endpoints will be searched (computationally expensive)'
    )
    print(usage_str.format(program=sys.argv[0]),)


if __name__ == '__main__':
    endpoints, output, interlinking = get_options(sys.argv[1:])
    create_rdfmts(endpoints, output, interlinking)



__author__ = "Philipp D. Rohde"

import re, sys
import time
from multiprocessing import Queue

from DeTrusty.Decomposer import Decomposer, Planner
from DeTrusty.Molecule.MTManager import ConfigFile
from DeTrusty.Wrapper.RDFWrapper import contact_source


re_https = re.compile("https?://")


def run_query(query: str,
              decomposition_type: str = "STAR",
              sparql_one_dot_one: bool = False,
              config: ConfigFile = ConfigFile('./Config/rdfmts.json'),
              join_stars_locally: bool = True,
              print_result: bool = True,
              yasqe: bool = False):
    """Executes a SPARQL query over a federation of SPARQL endpoints.

    The SPARQL query is decomposed based on the specified decomposition type.
    DeTrusty identifies the possible sources for each sub-query using the
    metadata collected previously. If the query contains the SERVICE clause,
    DeTrusty executed the sub-query at the specified endpoint instead.

    Parameters
    ----------
    query : str
        The SPARQL query to be executed.
    decomposition_type : str, optional
        The decomposition type to be used for decomposing the query. Possible values
        are 'STAR' for a star-shaped decomposition, 'EG' for exclusive groups decomposition,
        and 'TRIPLE' for a triple-wise decomposition, i.e., each triple pattern of the query
        produces a sub-query. Default is 'STAR'.
    sparql_one_dot_one : bool, optional
        Indicates whether the query includes the SERVICE clause.
        'True' meaning the SERVICE clause is present in the query, 'False' otherwise.
        Default is 'False'.
    config : DeTrusty.Molecule.MTManager.ConfigFile, optional
        The configuration holding the metadata about the federation over which the
        SPARQL query should be executed. If no value is specified, DeTrusty will
        attempt to load the configuration from `./Config/rdfmts.json`.
    join_stars_locally : bool, optional
        Indicates whether joins should be performed at the query engine.
        'True' meaning joins will be performed in DeTrusty, 'False' leads to joins
        being executed in the sources if possible. Default behavior is join execution
        at the query engine level.
    print_result : bool, optional
        Indicates whether the actual query result should be returned.
        'True' meaning the result will be included in the answer.
        'False' only returns the metadata of the query result, like the cardinality.
        Default is 'True'.
    yasqe : bool, optional
        Indicates whether the SPARQL query was sent from the YASGUI interface of
        DeTrusty's Web interface. This is a workaround for YASQE not being able
        to show the query results when the validation data is included.
        Set to 'True' to omit the validation data. Default is 'False'.

    Returns
    -------
    dict
        A dictionary including the query answer and additional metadata following the SPARQL protocol.
        It returns an error message in the 'error' field if something went wrong. Other metadata might
        be omitted in that case.

    """
    start_time = time.time()
    decomposer = Decomposer(query, config,
                            decompType=decomposition_type,
                            joinstarslocally=join_stars_locally,
                            sparql_one_dot_one=sparql_one_dot_one)
    decomposed_query = decomposer.decompose()

    if decomposed_query is None:
        return {"results": {}, "error": "The query cannot be answered by the endpoints in the federation."}

    planner = Planner(decomposed_query, True, contact_source, 'RDF', config)
    plan = planner.createPlan()

    output = Queue()
    plan.execute(output)

    result = []
    r = output.get()
    card = 0
    while r != 'EOF':
        card += 1
        if print_result:
            res = {}
            for key, value in r.items():
                res[key] = {"value": value, "type": "uri" if re_https.match(value) else "literal"}
            if not yasqe:
                res['__meta__'] = {"is_verified": True}

            result.append(res)
        r = output.get()
    end_time = time.time()

    return {"head": {"vars": decomposed_query.variables()},
            "cardinality": card,
            "results": {"bindings": result} if print_result else "printing results was disabled",
            "execution_time": end_time - start_time,
            "output_version": "2.0"}



from DeTrusty.Operators.AnapsidOperators.Xvalues import Xvalues
from DeTrusty.Sparql.Parser.services import Argument, Values
import time, unittest

class filterByValues(unittest.TestCase):

    def setUp(self):
        self.vars1 = Values([Argument('?disease_name')], [[Argument("\"Attention-deficit_hyperactivity_disorder\"", True)]])
        self.vars2 = Values([Argument('?disease_name')], [[None, True]])
        self.value1 = Xvalues(self.vars1)
        self.value2 = Xvalues(self.vars2)

    def test_isValid(self):
        tuple = {'disease_name': 'Attention-deficit_hyperactivity_disorder'}
        self.assertEqual(self.value1.filterByValues(tuple), True)

    def test_UNDEF(self):
        tuple = {'disease_name': 'Rheumatoid_arthritis'}
        self.assertEqual(self.value2.filterByValues(tuple), True)
    
    def test_NotValid(self):
        tuple = {'disease_name': 'Rheumatoid_arthritis'}
        self.assertEqual(self.value1.filterByValues(tuple), False)
    
    def tearDown(self):
        self.vars1 = None
        self.vars2 = None
        self.value1 = None
        self.value2 = None

class TestEfficiency(unittest.TestCase):

    def setUp(self):
        self.vars1 = Values([Argument("?disease_name")], [[Argument("\"Attention-deficit_hyperactivity_disorder\"", True)]])
        self._filterByValues = Xvalues(self.vars1)
        self._efficiency_data = dict()

    def test_filterByValues_single_var(self):
        starting_time = time.time()

        tuple = {'disease_name': 'Rheumatoid_arthritis'}
        self._filterByValues.filterByValues(tuple)

        ending_time = time.time()

        self._efficiency_data["filterByValues1"] = ending_time - starting_time

    def test_filterByValues_mult_var(self):
        starting_time = time.time()

        tuple = {'drug_gn': 'Heroin', 'drug': 'http://www4.wiwiss.fu-berlin.de/drugbank/resource/drugs/DB01452', 'disease_name': 'Response to morphine-6-glucuronide', 'disease': 'http://www4.wiwiss.fu-berlin.de/diseasome/resource/diseases/3733', 'disease_class': 'http://www4.wiwiss.fu-berlin.de/diseasome/resource/diseaseClass/Neurological'}
        self._filterByValues.filterByValues(tuple)

        ending_time = time.time()

        self._efficiency_data["filterByValues2"] = ending_time - starting_time
    
    def tearDown(self):
        print(self._efficiency_data)
        self.vars1 = None
        self._filterByValues = None

if __name__ == "__main__":
    unittest.main()



"""
Created on 10/12/2021.

Implements the Xhaving operator.
The intermediate results are represented in a queue.

@author: Avellino
"""

from multiprocessing import Queue
from DeTrusty.Sparql.Parser.services import HavingHelper

class Xhaving(object):

    name = "HAVING"

    def __init__(self, having):
        self.qresults = Queue()                                         # end res.
        self.having = having                                            # HAVING vars
    
    def execute(self, left, dummy, out, processqueue=Queue()):
        self.left = left
        self.qresults = out
        tuple = self.left.get(True)

        while tuple != "EOF":
            valid = execute_logic(tuple, self.having)
            if valid:
                self.qresults.put(tuple)
            tuple = self.left.get(True)
        self.qresults.put("EOF")
        return

def execute_logic(tup, constrain):
    bool_ls = list()

    for arg in constrain.args:
        if type(arg) is HavingHelper:
            bool_ls.append(execute_single(tup, arg))
        else:
            bool_ls.append(execute_logic(tup, arg))

    if constrain.log_op == 'AND':
        for el in bool_ls:
            if el is False:
                return False
        return True
    else:
        for el in bool_ls:
            if el is True:
                return True
        return False

def execute_single(tup, constrain):

            if constrain.op == 'EQUALS':
                if float(tup[constrain.agg.getName()]) == float(constrain.num):
                    return True
                return False 
                    
            elif constrain.op == 'NEQUALS':
                if float(tup[constrain.agg.getName()]) != float(constrain.num):
                    return True
                return False 

            elif constrain.op == 'GREATER':
                if float(tup[constrain.agg.getName()]) > float(constrain.num):
                    return True
                return False
            
            elif constrain.op == 'GREATEREQ':
                if float(tup[constrain.agg.getName()]) >= float(constrain.num):
                    return True
                return False

            elif constrain.op == 'LESS':
                if float(tup[constrain.agg.getName()]) < float(constrain.num):
                    return True
                return False

            elif constrain.op == 'LESSEQ':
                if float(tup[constrain.agg.getName()]) <= float(constrain.num):
                    return True
                return False



"""
Created on Mar 03, 2014

Implements the Xfilter operator.
The intermediate results are represented in a queue.

@author: Maribel Acosta Deibe
"""

from multiprocessing import Queue
from DeTrusty.Sparql.Parser.services import Filter, Expression, Argument
import datetime
import operator

unary_operators = {
    '!': operator.not_,
    '+': '',
    '-': operator.neg,
    'bound': lambda a: len(a) > 0,
    'str': str
}

logical_connectives = {
    '||': operator.or_,
    '&&': operator.and_
}

arithmetic_operators = {
    '*': operator.mul,
    '/': operator.truediv,
    '+': operator.add,
    '-': operator.sub,
}

test_operators = {
    '=': operator.eq,
    '!=': operator.ne,
    '<': operator.lt,
    '>': operator.gt,
    '<=': operator.le,
    '>=': operator.ge
}

data_types = {
    'integer': (int, 'numerical'),
    'decimal': (float, 'numerical'),
    'float': (float, 'numerical'),
    'double': (float, 'numerical'),
    'string': (str, str),
    'boolean': (bool, bool),
    'dateTime': (datetime, datetime),
    'nonPositiveInteger': (int, 'numerical'),
    'negativeInteger': (int, 'numerical'),
    'long': (int, 'numerical'),
    'int': (int, 'numerical'),
    'short': (int, 'numerical'),
    'byte': (bytes, bytes),
    'nonNegativeInteger': (int, 'numerical'),
    'unsignedLong': (int, 'numerical'),
    'unsignedInt': (int, 'numerical'),
    'unsignedShort': (int, 'numerical'),
    'unsignedByte': (bytes, bytes),  # TODO: this is not correct
    'positiveInteger': (int, 'numerical')
}

numerical = (int, int, float)


class Xfilter(object):

    name = "FILTER"

    def __init__(self, filter):
        self.input = Queue()
        self.qresults = Queue()
        self.filter = filter

    def execute(self, left, dummy, out, processqueue=Queue()):
        # Executes the Xfilter.
        self.left = left
        self.qresults = out
        # print "self.filter.expr.op", self.filter.expr.op
        # print "self.filter.expr.left", self.filter.expr.left
        # print "self.filter.expr.right", self.filter.expr.right
        # Apply filter tuple by tuple.
        tuple = self.left.get(True)

        # print 'tuple ', tuple

        while (tuple != "EOF"):
            (res, _) = self.evaluateComplexExpression(tuple, self.filter.expr.op, (self.filter.expr.left, None),
                                                      (self.filter.expr.right, None))
            if res:
                self.qresults.put(tuple)

            tuple = self.left.get(True)
            # print 'tuple ', tuple

        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        # return

    def __repr__(self):
        return str(self.__class__) + ">>  FILTER (" + str(self.filter.expr.left) + " " + str(
            self.filter.expr.op) + " " + str(self.filter.expr.right) + ")"

    # Base case.
    def evaluateOperator(self, operator, expr_left, expr_right):
        # print "operator in Filter", operator, expr_left, expr_right, type(expr_left), type(expr_right)
        if (operator in logical_connectives):
            # print "Case: logical connectives"
            return self.evaluateLogicalConnective(operator, expr_left, expr_right)
        elif operator in arithmetic_operators:
            # print "Case: arithmetic operator", operator
            return self.evaluateAritmethic(operator, expr_left, expr_right)
        elif (operator in unary_operators):
            # print "Case: unary_operators"
            return self.evaluateUnaryOperator(operator, expr_left)
        elif (operator in test_operators):
            # print "Case: test"
            return self.evaluateTest(operator, expr_left, expr_right)

    # Inductive case.
    def evaluateComplexExpression(self, tuple, operator, left, right):
        (expr_left, type_left), (expr_right, type_right) = left, right
        # Case 1: Inductive case binary operator OP(Expr, Expr)
        res = None
        if isinstance(expr_left, Expression) and isinstance(expr_right, Expression):
            # print "Case 1"
            res_left = self.evaluateComplexExpression(tuple, expr_left.op, (expr_left.left, type_left),
                                                      (expr_left.right, type_right))
            res_right = self.evaluateComplexExpression(tuple, expr_right.op, (expr_right.left, type_left),
                                                       (expr_right.right, type_right))

            res = self.evaluateOperator(operator, res_left, res_right)
            # print "res:", res
            # print "\n--------------------------"

        # Case 2: Inductive case binary operator OP(Expr, Arg)
        elif isinstance(expr_left, Expression) and isinstance(expr_right, Argument):
            # print "Case 2"
            res_left = self.evaluateComplexExpression(tuple, expr_left.op, (expr_left.left, type_left),
                                                      (expr_left.right, type_right))
            if expr_right.constant:
                res_right = expr_right.name
            else:
                res_right = self.extractValue(tuple[expr_right.name[1:]])
            res = self.evaluateOperator(operator, res_left, res_right)

        # Case 3: Inductive case binary operator OP(Arg, Expr)
        elif isinstance(expr_left, Argument) and isinstance(expr_right, Expression):
            # print "Case 3"
            # if expr_left.name[1:] in tuple:
            if expr_left.constant:
                res_left = (expr_left.name, str)
            else:
                res_left = self.extractValue(tuple[expr_left.name[1:]])
            res_right = self.evaluateComplexExpression(tuple, expr_right.op, (expr_right.left, type_left),
                                                       (expr_right.right, type_right))
            res = self.evaluateOperator(operator, res_left, res_right)
            # else:
            #   return None
        # Case 4: Inductive case unary operator OP(Expr, None)
        elif isinstance(expr_left, Expression):
            # print "Case 4"
            res_left = self.evaluateComplexExpression(tuple, expr_left.op, (expr_left.left, type_left),
                                                      (expr_left.right, type_right))
            res = self.evaluateOperator(operator, res_left, None)

        # Case 5: Base case binary operator OP(Arg, Arg)
        elif isinstance(expr_left, Argument) and isinstance(expr_right, Argument):
            # print "Case 5", expr_left.constant, expr_right.constant, tuple[expr_left.name[1:]], expr_right.name
            if expr_left.constant:
                res_left = expr_left.name, str
            else:
                res_left = self.extractValue(tuple[expr_left.name[1:]])
            if expr_right.constant:
                res_right = (expr_right.name, str)
            else:
                res_right = self.extractValue(tuple[expr_right.name[1:]])
            # print "res left, right ", res_left, operator, res_right
            res = self.evaluateOperator(operator, res_left, res_right)
            # print res

        # Case 6: Base case unary operator OP(Arg, None)
        elif isinstance(expr_left, Argument):
            # print "Case 6"
            if expr_left.constant:
                res_left = expr_left.name
            else:
                res_left = self.extractValue(tuple[expr_left.name[1:]])
            res = self.evaluateOperator(operator, res_left, None)
        else:
            pass
        return res

    '''
    evaluateEBV: calculates whether an argument is an Effective Boolean Value (EBV)
                 according to the definition in the SPARQL documentation 
                 See: http://www.w3.org/TR/sparql11-query/#ebv

    input: val -- an argument
    return: (isEBV, EBV) -- both of Python type bool
    '''

    def evaluateEBV(self, casted_val, type_val):

        # Handles python data types.
        if (isinstance(casted_val, bool)):
            return (True, casted_val)
        if (isinstance(casted_val, numerical)):
            if (casted_val == 0 or casted_val == 'nan'):
                return (True, False)
            else:
                return (True, True)

        # Rule 1
        if ((type_val == bool) and (casted_val != 'true') and (casted_val != 'false')):
            return (True, False)
        elif ((type_val == 'numeric') and not (isinstance(casted_val, numerical))):
            return (True, False)

        # Rule 2
        if (type_val == bool):
            if (casted_val == 'true'):
                return (True, True)
            elif (casted_val == 'false'):
                return (True, False)

        # Rule 3
        if (type_val == str):
            if (len(casted_val) == 0):
                return (True, False)
            else:
                return (True, True)

        # Rule 4
        if ((type_val == 'numeric')):
            if (casted_val == 0 or casted_val == 'nan'):
                return (True, False)
            else:
                return (True, True)

        # Rule 5: The error type should be raised by the evaluators.
        return (False, None)

    def evaluateUnaryOperator(self, operator, left):
        (expr_left, type_left) = left

        if (operator == '+' and isinstance(expr_left, numerical)):
            return (expr_left, type_left)

        elif (operator == '-' and isinstance(expr_left, numerical)):
            return (unary_operators[operator](expr_left), type_left)
        elif (operator == 'bound'):
            return (unary_operators[operator](expr_left), type_left)
        elif (operator == 'str'):
            return (unary_operators[operator](expr_left), str)
        elif (operator == '!'):
            (isEBV, ebv) = self.evaluateEBV(expr_left, type_left)
            if (isEBV):
                return (unary_operators[operator](ebv), type_left)
            else:
                raise SPARQLTypeError
        else:
            raise SPARQLTypeError

    def evaluateLogicalConnective(self, operator, left, right):
        (expr_left, type_left), (expr_right, type_right) = left, right

        (isEBV_left, ebv_left) = self.evaluateEBV(expr_left, type_left)
        (isEBV_right, ebv_right) = self.evaluateEBV(expr_right, type_right)

        # print "in evaluateLogicalConnective", expr_left, isEBV_left, ebv_left
        # print "in evaluateLogicalConnective", expr_right, isEBV_right, ebv_right

        if (isEBV_left and isEBV_right):
            return (logical_connectives[operator](ebv_left, ebv_right), bool)

        elif (isEBV_left):
            res = logical_connectives[operator](ebv_left, 'Error')
            if (res == 'Error'):
                raise SPARQLTypeError
            else:
                return (res, bool)

        elif (isEBV_right):
            res = logical_connectives[operator](ebv_right, 'Error')
            if (res == 'Error'):
                raise SPARQLTypeError
            else:
                return (res, bool)

    def evaluateTest(self, operator, left, right):
        if left and right:
            (expr_left, type_left), (expr_right, type_right) = left, right
        else:
            print("Exception: left or right operand None.", left, right)
            raise Exception
        if ((type(expr_left) == type(expr_right)) or (
                isinstance(expr_left, numerical) and isinstance(expr_right, numerical))):

            return (test_operators[operator](expr_left, expr_right), bool)
        else:
            try:
                if isinstance(expr_left, numerical):
                    ltyp = type(expr_left)
                    expr_right = ltyp(expr_right)
                elif isinstance(expr_right, numerical):
                    rtype = type(expr_right)
                    expr_left = rtype(expr_left)
                return (test_operators[operator](expr_left, expr_right), bool)
            except:
                print("SPARQLTypeError - in Xfilter")
                raise SPARQLTypeError

    def evaluateAritmethic(self, operator, left, right):
        (expr_left, type_left), (expr_right, type_right) = left, right
        # print "evaluateAritmethic(), ", expr_left, expr_right, operator, type(expr_left), type(expr_right), expr_right, type_left, type_right

        if (isinstance(expr_left, numerical) and isinstance(expr_right, numerical)):
            return (
            arithmetic_operators[operator](expr_left, expr_right), type_left)  # TODO: implement the cases with types
        elif (isinstance(expr_left, numerical) and isinstance(expr_right, str)):
            expr_right = int(expr_right)
            expr_left = int(expr_left)
            # print expr_right,expr_left , (arithmetic_operators[operator](expr_left, expr_right), type_left)
            return (
            arithmetic_operators[operator](expr_left, expr_right), type_left)  # TODO: implement the cases with types
        else:
            raise SPARQLTypeError

    def extractValue(self, val):
        pos = val.find("^^")
        # Handles when the literal is typed.
        if (pos > -1):
            for t in data_types.keys():
                if (t in val[pos:]):
                    (python_type, general_type) = data_types[t]
                    if (general_type == bool):
                        return (val[:pos], general_type)
                    else:
                        return (python_type(val[:pos]), general_type)
        else:
            return (str(val), str)


class SPARQLTypeError(Exception):
    """Base class for exceptions in this module."""
    pass



"""
Created on Jul 10, 2011

Implements a Hash Join operator.
The intermediate results are represented as a Queue, but it doesn't
work until all the tuples are arrived.

@author: Maribel Acosta Deibe
"""

from time import time
from .OperatorStructures import Table, Record
from DeTrusty.Operators.Join import Join
from multiprocessing import Queue


class HashJoin(Join):

    def __init__(self, vars):
        self.left_table  = Table()
        self.right_table = Table()
        self.results     = []
        self.vars        = vars

    def instantiate(self, d):
        newvars = self.vars - set(d.keys())
        return HashJoin(newvars)

    def execute(self, qleft, qright, out, processqueue=Queue()):
        # Executes the Hash Join.
        self.left = []
        self.right = []
        self.qresults = out

        # Initialize tuples.
        tuple1 = None
        tuple2 = None

        # Get the tuples from the queues.
        while (not(tuple1 == "EOF") or not(tuple2 == "EOF")):
            # Try to get tuple from left queue.
            if not(tuple1 == "EOF"):
                try:
                    tuple1 = qleft.get(False)
                    #print tuple1
                    self.left.append(tuple1)
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.left.get(False), when the queue is empty.
                    pass

            # Try to get tuple from right queue.
            if not(tuple2 == "EOF"):
                try:
                    tuple2 = qright.get(False)
                    #print tuple2
                    self.right.append(tuple2)
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.right.get(False), when the queue is empty.
                    pass

        # Get the variables to join.
        if ((len(self.left) > 1) and (len(self.right) > 1)):
            # Iterate over the lists to get the tuples.
            while ((len(self.left) > 1) or (len(self.right) > 1)):
                if len(self.left) > 1:
                    self.insertAndProbe(self.left.pop(0), self.left_table, self.right_table)
                if len(self.right) > 1:
                    self.insertAndProbe(self.right.pop(0), self.right_table, self.left_table)

        # Put all the results in the output queue.
        while self.results:
            self.qresults.put(self.results.pop(0))

        # Put EOF in queue and exit.
        self.qresults.put("EOF")

    def insertAndProbe(self, tuple, table1, table2):
        # Insert the tuple in its corresponding partition and probe.
        #print tuple
        # Get the attribute(s) to apply hash.
        att = ''
        for var in self.vars:
            att = att + tuple[var]
        i = hash(att) % table1.size

        # Insert record in partition.
        record = Record(tuple, time(), 0)
        table1.insertRecord(i, record)

        # Probe the record against its partition in the other table.
        self.probe(record, table2.partitions[i], self.vars)

    def probe(self, record, partition, var):
        # Probe a tuple if the partition is not empty.
        if partition:

            # For every record in the partition, check if it is duplicated.
            # Then, check if the tuple matches for every join variable.
            # If there is a join, concatenate the tuples and produce result.
            for r in partition.records:
                if self.isDuplicated(record, r):
                    break
                join = True
                for v in var:
                    join = True
                    if record.tuple[v] != r.tuple[v]:
                        join = False
                        break

                if join:
                    res = record.tuple.copy()
                    res.update(r.tuple)
                    self.results.append(res)

    def isDuplicated(self, record1, record2):
        # Verify if the tuples has been already probed.
        return not record1.ats >= record2.ats



#!/usr/bin/env python3
import setuptools

with open("VERSION", "r", encoding="utf-8") as ver:
    version = ver.read()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='DeTrusty',
    version=version,
    description='DeTrusty - Decentralized and Trustable Query Engine',
    license='GNU/GPLv3',
    author='Philipp D. Rohde',
    author_email='philipp.rohde@tib.eu',
    url='https://github.com/SDM-TIB/DeTrusty',
    download_url='https://github.com/SDM-TIB/DeTrusty/archive/refs/tags/v' + version + '.tar.gz',
    long_description=long_description,
    long_description_content_type="text/markdown",
    scripts=['./Scripts/create_rdfmts.py',
             './Scripts/restart_workers.sh'],
    packages=[
        'DeTrusty',
        'DeTrusty.Decomposer',
        'DeTrusty.Molecule',
        'DeTrusty.Operators',
        'DeTrusty.Operators.AnapsidOperators',
        'DeTrusty.Operators.BlockingOperators',
        'DeTrusty.Operators.NonBlockingOperators',
        'DeTrusty.Sparql',
        'DeTrusty.Sparql.Parser',
        'DeTrusty.Wrapper',
        'DeTrusty.Wrapper.RDFWrapper'
    ],
    install_requires=['requests>=2.27.0',
                      'ply==3.11',
                      'rdflib>=6.0.0'],
    include_package_data=True,
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research'
    ]
)



from DeTrusty.Sparql.Parser.services import Aggregate, HavingHelper, Expression

def getVars(sg):
    s = []
    if not sg.subject.constant:
        s.append(sg.subject.name)
    if not sg.theobject.constant:
        s.append(sg.theobject.name)
    return s

def getPrefs(ps):
    prefDict = dict()
    for p in ps:
         pos = p.find(":")
         c = p[0:pos].strip()
         v = p[(pos+1):len(p)].strip()
         prefDict[c] = v
    return prefDict


def getUri(p, prefs):
    if "'" in p.name or '"' in p.name:
        return p.name
    hasPrefix = prefix(p)
    if hasPrefix:
        (pr, su) = hasPrefix
        n = prefs[pr]
        n = n[:-1] + su + ">"
        return n
    return p.name


def prefix(p):
    s = p.name
    pos = s.find(":")
    if (not (s[0] == "<")) and pos > -1:
        return (s[0:pos].strip(), s[(pos+1):].strip())

    return None


def getTemplatesPerMolecule(molecules):
    moltemp = dict()
    for m in molecules:
        for t in molecules[m].templates:
            if m in moltemp:
                moltemp[m].append(t.pred)
            else:
                moltemp[m] = [t.pred]
    return moltemp

###################################################################

def collectVars(proj, group, having, agg):
    implicit_group = not group 
    agg_exist = False
    over_all_triples = True
    index = 0

    for var in proj:
        if type(var) is Aggregate or type(var) is Expression:
            agg_exist = True
            break

    for var in proj:
        if type(var) is not Aggregate and type(var) is not Expression:
            over_all_triples = False
            if implicit_group and agg_exist:
                group.append(var)
        else:
            var.name = "?callret-" + str(index)
            agg.append(var)
        index += 1

    if having:
        tmp = havingVars(having.args)
        for arg in tmp:
            if arg not in agg:
                agg.append(arg)

    return over_all_triples

def havingVars(tmp):
    ret = list()
    for x in tmp:
        if type(x) is HavingHelper:
            ret.append(x.agg) 
        else:
            ret += havingVars(x.args)
    return ret
